//
// This sketch implements a "Jeopardy" style contestant response tiebreaker.
//
// Each contestant has a pushbutton switch.
// After setup, the waitingForQuestion function is invoked. This uses a
// local loop, rather than the Arduino loop, to wait for the moderator
// to begin. While waiting, the player lights are illuminated in chase mode.

// When the moderator presses the switch S6 button, the QUESTION_READING
// state begins. There is no time limit here. In this phase the player
// LEDs will light in chase mode. During this time, if a player button is
// pressed, a penalty of 0.25 seconds is applied for the next phase. When
// the moderator presses the switch S5 button and releases it, the
// READY_FOR_ANSWER_LIGHT is turned on, player LEDs will be extinguished and
// the game will proceed to the WAITING_FOR_RING_IN state.

// In WAITING_FOR_RING_IN, the READY_FOR_ANSWER_LIGHT is
// illuminated. There is a five second timeout which, at its expiration,
// causes the game to resetAll to clear variables, the waitingForQuestion
// function is invoked and the state changes to QUESTION_READING. When
// a player presses a button, if the player is not BLOCKED from a prior
// wrong answer and any delay time imposed in the QUESTION_READING phase,
// that player's light is turned and READY_FOR_ANSWER_LIGHT is turned off.
// The game proceeds to the WAITING_FOR_ANSWER state.

// In WAITING_FOR_ANSWER, after a five second timeout or
// a player giving a wrong answer, the game proceeds to
// WAITING_FOR_RING_IN, if there are players not yet buzzed in. If a
// player gives a correct answer or all players have attempted to answer,
// resetAll clears variables, the waitingForQuestion function is invoked
// and the state changes to QUESTION_READING.
//

// Note: This program is set up to handle up to four players. If there are fewer
// players, it will handle them, although when those players have rung in and
// incorrectly answered, the program will time out, instead of immediately
// recognizing that there are no remaining players. For the traditional three
// player purists or those who cannot wait the extra five seconds, MAX_PLAYERS
// can be changed to three and the program recompilied and uploaded. It will then
// only recognize players one through three.

// 2012.12.18 Initial release.

/*
 * Environment definitions
 */
#define MAX_PLAYERS 4
#define TIMEOUT_PERIOD 5000
#define DELAY_TIME 250
#define NO_ANSWER 0
#define CORRECT_ANSWER 1
#define WRONG_ANSWER -1

/*
 * Hardware definitions
  */
#define TOP_LED 13
#define TOP_SWITCH 9
// The CORRECT_ANSWER_BUTTON is switch S6
#define CORRECT_ANSWER_BUTTON 3
// The WRONG_ANSWER_BUTTON is switch S5
#define WRONG_ANSWER_BUTTON 4
#define READY_FOR_ANSWER_LIGHT 5

/*
 * Housekeeping definitions
 */
int buzzedInPlayer = -1;
int buzzedInPlayerCount = 0;
unsigned long timeout = 0;
int blockedPlayers = 0;
int moderator;

char ESC_CHAR = 27;
#define PUT_CURSOR_MIDSCREEN\
    Serial.print(ESC_CHAR);\
    Serial.print("[2J");\
    delay(2);\
    Serial.print(ESC_CHAR);\
    Serial.print("[10;1H");


/*
 *  Button press definitions
 */
#define NOT_PUSHED HIGH
#define PUSHED     LOW
#define BLOCKED   -1

/*
 *  LED state definitions
 */
#define LED_ON  HIGH
#define LED_OFF LOW

/*
 * Game States
 */
#define QUESTION_READING 1
#define WAITING_FOR_RING_IN 2
#define WAITING_FOR_ANSWER 3
int phase = QUESTION_READING;
#define DEBOUNCE_TIME 10

struct player_t {
    int light;
    int buttonPin;
    int button;
    unsigned timePenalty;
} ;// End struct playerStruct_t

struct player_t player[MAX_PLAYERS];

// Reset player time.
void resetPlayerTime()
{
   int i;

   for (i = 0; i < MAX_PLAYERS; i++)
   {
       player[i].timePenalty = 0;
   };
} // End resetAll

// Reset to the initial state.
void resetAll()
{
   int i;
   moderator = NO_ANSWER;

   phase = QUESTION_READING;

   for (i = 0; i < MAX_PLAYERS; i++)
   {
       digitalWrite(player[i].light, LED_OFF);
       player[i].button = NOT_PUSHED;
       resetPlayerTime();
   };
   buzzedInPlayer = -1;
   buzzedInPlayerCount = 0;
} // End resetAll

int readModerator()
{
// Button release is the trigger for a button having been pushed.
// Debounce is used to sense the button release.
    if (digitalRead(CORRECT_ANSWER_BUTTON) == PUSHED) {
        for(;;) {
            delay(DEBOUNCE_TIME);
            if (digitalRead(CORRECT_ANSWER_BUTTON) == NOT_PUSHED) {
                break;
            }
        }
        return CORRECT_ANSWER;
    } else if (digitalRead(WRONG_ANSWER_BUTTON) == PUSHED) {
        for(;;) {
            delay(DEBOUNCE_TIME);
            if (digitalRead(WRONG_ANSWER_BUTTON) == NOT_PUSHED) {
                break;
            }
        }
        return WRONG_ANSWER;
    } else {
        return NO_ANSWER;
    }
}

void readPlayers()
{
// States of all player buttons are saved into the player array.
    int i;
    for (i = 0; i < MAX_PLAYERS; i++)
    {
        if (player[i].button != BLOCKED) {
            if (digitalRead(player[i].buttonPin) == PUSHED) {
                player[i].button = PUSHED;
            } else {
                player[i].button = NOT_PUSHED;
            }
        }
    }
}

void waitingForQuestion()
{
    int i;

    digitalWrite(READY_FOR_ANSWER_LIGHT, LED_OFF);
    Serial.println("\tModerator, Press S6 and begin reading");
    Serial.println("\tModerator, Press S5 when you have finished the Question");
    while (!(readModerator() == CORRECT_ANSWER)) {
        for (i = 0; i < MAX_PLAYERS; i++) {
            digitalWrite(player[i].light, LED_ON);
            delay(250);
            digitalWrite(player[i].light, LED_OFF);
        }
    }
    PUT_CURSOR_MIDSCREEN
    Serial.println("\tPlayers, Wait for question to finish");
}
void setup()
{
    int i;
    for (i = 0; i < MAX_PLAYERS; i++) {

        // Assign the LEDs, cathodes connected to ground
        player[i].light = TOP_LED - i;
        pinMode(player[i].light, OUTPUT);

        // Assign the pushbutton switches; connect the input to ground when closed.
        player[i].buttonPin = TOP_SWITCH - i;
        pinMode(player[i].buttonPin, INPUT);
        digitalWrite(player[i].buttonPin, HIGH);  // Enable pullup

    }

    pinMode(READY_FOR_ANSWER_LIGHT, OUTPUT);

    pinMode( WRONG_ANSWER_BUTTON, INPUT );
    digitalWrite( WRONG_ANSWER_BUTTON, HIGH);    // Enable pullup

    pinMode( CORRECT_ANSWER_BUTTON, INPUT );
    digitalWrite( CORRECT_ANSWER_BUTTON, HIGH);    // Enable pullup

    resetAll();

    Serial.begin(9600);           // set up Serial library at 9600 bps

    PUT_CURSOR_MIDSCREEN
    waitingForQuestion();
}

void loop()
{
    int i;

    switch (phase) {
    case QUESTION_READING: {
    // Moderator has pressed the reset button
    // or the questioned has been answeredd
    // or all players have answered incorrectly.
    // No timing for this phase.
    // Any Buzz-ins will suffer 0.25 second delay when next state entered.
        readPlayers();
        // Apply Penalty for eager button pressers
        for (i = 0; i < MAX_PLAYERS; i++) {
            if ((player[i].button == PUSHED) &&
                (player[i].timePenalty == 0)) {
                player[i].timePenalty = DELAY_TIME;
                Serial.print("\tPlayer ");
                Serial.print(i+1);
                Serial.println(" gets a penalty");
            }
        }
        moderator = readModerator();
        if (moderator == WRONG_ANSWER) {
            phase = WAITING_FOR_RING_IN;
            timeout = millis() + TIMEOUT_PERIOD;
            digitalWrite(READY_FOR_ANSWER_LIGHT, LED_ON);
            for (i = 0; i < MAX_PLAYERS; i++) {
                player[i].timePenalty += millis();
            }
            PUT_CURSOR_MIDSCREEN
            Serial.println("\tWaiting for Ring In");
            Serial.println();
            blockedPlayers = 0;
        }
        break;
    } // End case QUESTION_READING

    case WAITING_FOR_RING_IN: {
    // Moderator has pressed the ANY BUTTON while in QUESTION_READING state.
    // Five seconds for this phase.
        if (blockedPlayers >= MAX_PLAYERS) {
            PUT_CURSOR_MIDSCREEN
            Serial.println("\tAll players have rung in");
            Serial.println("\tModerator, Please read the next Question");
            phase = QUESTION_READING;
            resetAll();
            waitingForQuestion();
            break;
        } else {
            buzzedInPlayer = -1;
            if (timeout >= millis()) {
                readPlayers();
                for (i = 0; i < MAX_PLAYERS; i++) {
                    if ((player[i].timePenalty <= millis()) &&
                        player[i].button == PUSHED) {
                        buzzedInPlayer = i;
                        buzzedInPlayerCount += 1;
                        digitalWrite(player[buzzedInPlayer].light, LED_ON);
                        digitalWrite(READY_FOR_ANSWER_LIGHT, LED_OFF);
                        phase = WAITING_FOR_ANSWER;
                        timeout = millis() + TIMEOUT_PERIOD;
                        player[i].button = BLOCKED;
                        blockedPlayers += 1;
                        PUT_CURSOR_MIDSCREEN
                        Serial.print("\tPlayer Number ");
                        Serial.println(i+1);
                    }
                }
            } else {
                Serial.println("\tTimed out waiting for a RING_IN ");
                Serial.println("\tModerator, Please read the next Question");
                phase = QUESTION_READING;
                resetAll();
                waitingForQuestion();
                break;
            }
            break;
        } // End else
    } // End case WAITING_FOR_RING_IN

    case WAITING_FOR_ANSWER: {
    // A player has PUSHED the button while in WAITING_FOR_RING_IN state.
    // Five seconds for this phase.
        if (timeout < millis()) {
            if (buzzedInPlayerCount <= MAX_PLAYERS) {
                PUT_CURSOR_MIDSCREEN
                Serial.println("\tWAITING_FOR_RING_IN");
                timeout = millis() + TIMEOUT_PERIOD;
                digitalWrite(player[buzzedInPlayer].light, LED_OFF);
                digitalWrite(READY_FOR_ANSWER_LIGHT, LED_ON);
                phase = WAITING_FOR_RING_IN;
                PUT_CURSOR_MIDSCREEN
                Serial.print("\tPlayer Number ");
                Serial.print(buzzedInPlayer+1);
                Serial.println(" timed out waiting for an answer ");
                }
        }
        moderator = readModerator();
        if (moderator == CORRECT_ANSWER) {
            digitalWrite(player[buzzedInPlayer].light, LED_OFF);
            PUT_CURSOR_MIDSCREEN
            Serial.print("\tPlayer Number ");
            Serial.print(buzzedInPlayer+1);
            Serial.println(", that is correct");
            Serial.println("\tModerator, Please read the next Question");
            phase = QUESTION_READING;
            resetAll();
            waitingForQuestion();
        } else
        if (moderator == WRONG_ANSWER) {
            phase = WAITING_FOR_RING_IN;
            timeout = millis() + TIMEOUT_PERIOD;
            digitalWrite(player[buzzedInPlayer].light, LED_OFF);
            digitalWrite(READY_FOR_ANSWER_LIGHT, LED_ON);
            Serial.println("\tOooo, sorry, that is wrong");
            Serial.println("\tWaiting for Ring In");
            resetPlayerTime();
        }
        break;
    } // End case WAITING_FOR_ANSWER

    default: {
        Serial.print("default: phase = ");
        Serial.println(phase);
    }

     } // End switch (phase)
}
